@@ -33,6 +33,7 @@
 #include <errno.h>
 #include <unistd.h>
 #include <sys/time.h>
+#include <sys/wait.h>
 
 #include "ratpoison.h"
 
@@ -44,7 +45,7 @@ static void
 new_window (XCreateWindowEvent *e)
 {
   rp_window *win;
-  screen_info *s;
+  rp_screen *s;
 
   if (e->override_redirect)
     return;
@@ -64,40 +65,10 @@ new_window (XCreateWindowEvent *e)
     }
 }
 
-static void
-cleanup_frame (rp_window_frame *frame)
-{
-  rp_window *last_win;
-  rp_window *win;
-
-  win = find_window_other ();
-  if (win == NULL)
-    {
-      set_frames_window (frame, NULL);
-      return;
-    }
-
-  last_win = set_frames_window (frame, win);
-
-  maximize (win);
-  unhide_window (win);
-
-
-#ifdef MAXSIZE_WINDOWS_ARE_TRANSIENTS
-  if (!win->transient
-      && !(win->hints->flags & PMaxSize 
-	   && win->hints->max_width < win->scr->root_attr.width
-	   && win->hints->max_height < win->scr->root_attr.height))
-#else
-  if (!win->transient)
-#endif
-    hide_others (win);
-}
-
 static void 
 unmap_notify (XEvent *ev)
 {
-  rp_window_frame *frame;
+  rp_frame *frame;
   rp_window *win;
 
   /* ignore SubstructureNotify unmaps. */
@@ -106,7 +77,7 @@ unmap_notify (XEvent *ev)
     return;
 
   /* FIXME: Should we only look in the mapped window list? */
-  win = find_window_in_list (ev->xunmap.window, rp_mapped_window_sentinel);
+  win = find_window_in_list (ev->xunmap.window, &rp_mapped_window);
 
   if (win == NULL)
     return;
@@ -114,16 +85,19 @@ unmap_notify (XEvent *ev)
   switch (win->state)
     {
     case IconicState:
-      PRINT_DEBUG ("Withdrawing iconized window '%s'\n", window_name (win));
+      PRINT_DEBUG (("Withdrawing iconized window '%s'\n", window_name (win)));
       if (ev->xunmap.send_event) withdraw_window (win);
       break;
     case NormalState:
-      PRINT_DEBUG ("Withdrawing normal window '%s'\n", window_name (win));
+      PRINT_DEBUG (("Withdrawing normal window '%s'\n", window_name (win)));
       /* If the window was inside a frame, fill the frame with another
 	 window. */
       frame = find_windows_frame (win);
-      if (frame) cleanup_frame (frame);
-      if (frame == win->scr->rp_current_frame) set_active_frame (frame);
+      if (frame)
+	{
+	  cleanup_frame (frame);
+	  if (frame->number == win->scr->current_frame) set_active_frame (frame);
+	}
 
       withdraw_window (win);
       break;
@@ -140,31 +114,31 @@ map_request (XEvent *ev)
   win = find_window (ev->xmap.window);
   if (win == NULL) 
     {
-      PRINT_DEBUG ("Map request from an unknown window.\n");
+      PRINT_DEBUG (("Map request from an unknown window.\n"));
       XMapWindow (dpy, ev->xmap.window);
       return;
     }
 
-  PRINT_DEBUG ("Map request from a managed window\n");
+  PRINT_DEBUG (("Map request from a managed window\n"));
 
   switch (win->state)
     {
     case WithdrawnState:
       if (unmanaged_window (win->w))
 	{
-	  PRINT_DEBUG ("Mapping Unmanaged Window\n");
+	  PRINT_DEBUG (("Mapping Unmanaged Window\n"));
 	  XMapWindow (dpy, win->w);
 	  break;
 	}
       else
 	{
-	  PRINT_DEBUG ("Mapping Withdrawn Window\n");
+	  PRINT_DEBUG (("Mapping Withdrawn Window\n"));
 	  map_window (win);
 	  break;
 	}
       break;
     case IconicState:
-      PRINT_DEBUG ("Mapping Iconic window\n");
+      PRINT_DEBUG (("Mapping Iconic window\n"));
       if (win->last_access == 0)
 	{
 	  /* Depending on the rudeness level, actually map the
@@ -197,23 +171,13 @@ map_request (XEvent *ev)
 static void
 destroy_window (XDestroyWindowEvent *ev)
 {
-  rp_window_frame *frame;
   rp_window *win;
 
   win = find_window (ev->window);
   if (win == NULL) return;
 
   ignore_badwindow++;
-
-  /* A destroyed window should never have a frame, since it should
-     have been cleaned up with an unmap notify event, but just in
-     case... */
-  frame = find_windows_frame (win);
-  if (frame) cleanup_frame (frame);
-  if (frame == win->scr->rp_current_frame) set_active_frame (frame);
-
   unmanage (win);
-
   ignore_badwindow--;
 }
 
@@ -253,46 +217,46 @@ configure_request (XConfigureRequestEvent *e)
 
 	    }
 
-	  PRINT_DEBUG("request CWStackMode %d\n", e->detail);
+	  PRINT_DEBUG(("request CWStackMode %d\n", e->detail));
 	}
 
-      PRINT_DEBUG ("'%s' window size: %d %d %d %d %d\n", window_name (win),
-		   win->x, win->y, win->width, win->height, win->border);
+      PRINT_DEBUG (("'%s' window size: %d %d %d %d %d\n", window_name (win),
+		   win->x, win->y, win->width, win->height, win->border));
 
       /* Collect the changes to be granted. */
       if (e->value_mask & CWBorderWidth)
 	{
 	  changes.border_width = e->border_width;
 	  win->border = e->border_width;
-	  PRINT_DEBUG("request CWBorderWidth %d\n", e->border_width);
+	  PRINT_DEBUG(("request CWBorderWidth %d\n", e->border_width));
 	}
 
       if (e->value_mask & CWWidth)
 	{
 	  changes.width = e->width;
 	  win->width = e->width;
-	  PRINT_DEBUG("request CWWidth %d\n", e->width);
+	  PRINT_DEBUG(("request CWWidth %d\n", e->width));
 	}
 
       if (e->value_mask & CWHeight)
 	{
 	  changes.height = e->height;
 	  win->height = e->height;
-	  PRINT_DEBUG("request CWHeight %d\n", e->height);
+	  PRINT_DEBUG(("request CWHeight %d\n", e->height));
 	}
 
       if (e->value_mask & CWX)
 	{
 	  changes.x = e->x;
 	  win->x = e->x;
-	  PRINT_DEBUG("request CWX %d\n", e->x);
+	  PRINT_DEBUG(("request CWX %d\n", e->x));
 	}
 
       if (e->value_mask & CWY)
 	{
 	  changes.y = e->y;
 	  win->y = e->y;
-	  PRINT_DEBUG("request CWY %d\n", e->y);
+	  PRINT_DEBUG(("request CWY %d\n", e->y));
 	}
 
       if (e->value_mask & (CWX|CWY|CWBorderWidth|CWWidth|CWHeight))
@@ -324,13 +288,13 @@ configure_request (XConfigureRequestEvent *e)
 static void
 client_msg (XClientMessageEvent *ev)
 {
-  PRINT_DEBUG ("Received client message.\n");
+  PRINT_DEBUG (("Received client message.\n"));
 
   if (ev->message_type == wm_change_state)
     {
       rp_window *win;
 
-      PRINT_DEBUG ("WM_CHANGE_STATE\n")
+      PRINT_DEBUG (("WM_CHANGE_STATE\n"));
 
       win = find_window (ev->window);
       if (win == NULL) return;
@@ -340,7 +304,7 @@ client_msg (XClientMessageEvent *ev)
 	     user's intervention. This is bad, but Emacs is the only
 	     program I know of that iconifies itself and this is
 	     generally from the user pressing C-z.  */
-	  PRINT_DEBUG ("Iconify Request.\n");
+	  PRINT_DEBUG (("Iconify Request.\n"));
 	  if (win->state == NormalState)
 	    {
 	      rp_window *w = find_window_other();
@@ -348,12 +312,12 @@ client_msg (XClientMessageEvent *ev)
 	      if (w)
 		set_active_window (w);
 	      else
-		blank_frame (win->scr->rp_current_frame);
+		blank_frame (screen_get_frame (win->scr, win->scr->current_frame));
 	    }
 	}
       else
 	{
-	  PRINT_ERROR ("Non-standard WM_CHANGE_STATE format\n");
+	  PRINT_ERROR (("Non-standard WM_CHANGE_STATE format\n"));
 	}
     }
 }
@@ -373,7 +337,7 @@ ungrab_rat ()
 }
 
 static void
-handle_key (screen_info *s)
+handle_key (rp_screen *s)
 {
   char *keysym_name;
   rp_action *key_action;
@@ -383,7 +347,7 @@ handle_key (screen_info *s)
   unsigned int mod;		/* Modifiers */
   int rat_grabbed = 0;
 
-  PRINT_DEBUG ("handling key...\n");
+  PRINT_DEBUG (("handling key...\n"));
 
   /* All functions hide the program bar and the frame indicator. */
   if (defaults.bar_timeout > 0) hide_bar (s);
@@ -404,8 +368,11 @@ handle_key (screen_info *s)
       rat_grabbed = 1;
     }
 
-  read_key (&keysym, &mod, NULL, 0);
+  read_key (&keysym, &mod, NULL, 0, 1);
+
   XSetInputFocus (dpy, fwin, revert, CurrentTime);
+  if (rat_grabbed)
+    ungrab_rat();
 
   if ((key_action = find_keybinding (keysym, x11_mask_to_rp_mask (mod))))
     {
@@ -423,15 +390,12 @@ handle_key (screen_info *s)
       marked_message_printf (0, 0, " %s unbound key ", keysym_name);
       free (keysym_name);
     }
-
-  if (rat_grabbed)
-    ungrab_rat();
 }
 
 static void
 key_press (XEvent *ev)
 {
-  screen_info *s;
+  rp_screen *s;
   unsigned int modifier;
   KeySym ks;
 
@@ -470,6 +434,7 @@ key_press (XEvent *ev)
 static char *
 execute_remote_command (Window w)
 {
+  int status;
   char *result = NULL;
   Atom type_ret;
   int format_ret;
@@ -477,28 +442,36 @@ execute_remote_command (Window w)
   unsigned long bytes_after;
   unsigned char *req;
 
-  if (XGetWindowProperty (dpy, w, rp_command,
-			  0, 0, False, XA_STRING,
-			  &type_ret, &format_ret, &nitems, &bytes_after,
-			  &req) == Success
-      &&
-      XGetWindowProperty (dpy, w, rp_command,
-			  0, (bytes_after / 4) + (bytes_after % 4 ? 1 : 0),
-			  True, XA_STRING, &type_ret, &format_ret, &nitems, 
-			  &bytes_after, &req) == Success)
+  status = XGetWindowProperty (dpy, w, rp_command,
+			       0, 0, False, XA_STRING,
+			       &type_ret, &format_ret, &nitems, &bytes_after,
+			       &req);
+
+  if (status != Success || req == NULL)
     {
-      if (req)
-	{
-	  PRINT_DEBUG ("command: %s\n", req);
-	  result = command (0, req);
-	}
-      XFree (req);
+      PRINT_DEBUG (("Couldn't get RP_COMMAND Property\n"));
+      return NULL;
     }
-  else
+
+  /* XGetWindowProperty always allocates one extra byte even if
+     the property is zero length. */  
+  XFree (req);
+
+  status = XGetWindowProperty (dpy, w, rp_command,
+			       0, (bytes_after / 4) + (bytes_after % 4 ? 1 : 0),
+			       True, XA_STRING, &type_ret, &format_ret, &nitems, 
+			       &bytes_after, &req);
+
+  if (status != Success || req == NULL)
     {
-      PRINT_DEBUG ("Couldn't get RP_COMMAND Property\n");
+      PRINT_DEBUG (("Couldn't get RP_COMMAND Property\n"));
+      return NULL;
     }
 
+  PRINT_DEBUG (("command: %s\n", req));
+  result = command (0, req);
+  XFree (req);
+
   return result;
 }
 
@@ -516,44 +489,66 @@ receive_command ()
   unsigned long nitems;
   unsigned long bytes_after;
   void *prop_return;
+  int offset;
 
+  /* Init offset to 0. In the case where there is more than one window
+     in the property, a partial read does not delete the property and
+     we need to grab the next window by incementing offset to the
+     offset of the next window. */
+  offset = 0;
   do
     {
-      if (XGetWindowProperty (dpy, DefaultRootWindow (dpy),
-			      rp_command_request, 0, 
-			      sizeof (Window) / 4 + (sizeof (Window) % 4 ?1:0),
-			      True, XA_WINDOW, &type_ret, &format_ret, &nitems,
-			      &bytes_after, (unsigned char **)&prop_return) == Success)
+      int ret;
+      int length;
+      Window w;
+
+      length = sizeof (Window) / 4 + (sizeof (Window) % 4 ?1:0);
+      ret = XGetWindowProperty (dpy, DefaultRootWindow (dpy),
+				rp_command_request, 
+				offset, length, 
+				True, XA_WINDOW, &type_ret, &format_ret,
+				&nitems,
+				&bytes_after, (unsigned char **)&prop_return);
+
+      /* Update the offset to point to the next window (if there is
+	 another one). */
+      offset += length;
+
+      if (ret != Success)
 	{
+	  PRINT_ERROR (("XGetWindowProperty Failed\n"));
 	  if (prop_return)
-	    {
-	      Window w;
-
-	      w = *(Window *)prop_return;
-	      XFree (prop_return);
-
-	      result = execute_remote_command (w);
-	      if (result)
-		{
-		  XChangeProperty (dpy, w, rp_command_result, XA_STRING,
-				   8, PropModeReplace, result, strlen (result));
-		  free (result);
-		}
-	      else
-		{
-		  XChangeProperty (dpy, w, rp_command_result, XA_STRING,
-				   8, PropModeReplace, NULL, 0);
-		}
-	    }
-	  else
-	    {
-	      PRINT_DEBUG ("Couldn't get RP_COMMAND_REQUEST Property\n");
-	    }
-
-	  PRINT_DEBUG ("command requests: %ld\n", nitems);
+	    XFree (prop_return);
+	  break;
 	}
-    } while (nitems > 0);
-			  
+
+      /* If there was no window, then we're done. */
+      if (prop_return == NULL)
+	{
+	  PRINT_DEBUG (("No property to read\n"));
+	  break;
+	}
+
+      /* We grabbed a window, so now find read the command stored in
+	 this window and execute it. */
+      w = *(Window *)prop_return;
+      XFree (prop_return);
+      result = execute_remote_command (w);
+
+      /* notify the client of any text that was returned by the
+	 command. */
+      if (result)
+	{
+	  XChangeProperty (dpy, w, rp_command_result, XA_STRING,
+			   8, PropModeReplace, result, strlen (result));
+	  free (result);
+	}
+      else
+	{
+	  XChangeProperty (dpy, w, rp_command_result, XA_STRING,
+			   8, PropModeReplace, NULL, 0);
+	}
+    } while (bytes_after > 0);
 }
 
 static void
@@ -561,13 +556,13 @@ property_notify (XEvent *ev)
 {
   rp_window *win;
 
-  PRINT_DEBUG ("atom: %ld\n", ev->xproperty.atom);
+  PRINT_DEBUG (("atom: %ld\n", ev->xproperty.atom));
 
   if (ev->xproperty.atom == rp_command_request
       && ev->xproperty.window == DefaultRootWindow (dpy)
       && ev->xproperty.state == PropertyNewValue)
     {
-      PRINT_DEBUG ("ratpoison command\n");
+      PRINT_DEBUG (("ratpoison command\n"));
       receive_command();
     }
 
@@ -578,25 +573,25 @@ property_notify (XEvent *ev)
       switch (ev->xproperty.atom)
 	{
 	case XA_WM_NAME:
-	  PRINT_DEBUG ("updating window name\n");
+	  PRINT_DEBUG (("updating window name\n"));
 	  update_window_name (win);
 	  update_window_names (win->scr);
 	  break;
 
 	case XA_WM_NORMAL_HINTS:
-	  PRINT_DEBUG ("updating window normal hints\n");
+	  PRINT_DEBUG (("updating window normal hints\n"));
 	  update_normal_hints (win);
 	  if (win->state == NormalState)
 	    maximize (win);
 	  break;
 
 	case XA_WM_TRANSIENT_FOR:
-	  PRINT_DEBUG ("Transient for\n");
+	  PRINT_DEBUG (("Transient for\n"));
 	  win->transient = XGetTransientForHint (dpy, win->w, &win->transient_for);
 	  break;
 
 	default:
-	  PRINT_DEBUG ("Unhandled property notify event\n");
+	  PRINT_DEBUG (("Unhandled property notify event\n"));
 	  break;
 	}
     }
@@ -641,7 +636,7 @@ focus_change (XFocusChangeEvent *ev)
 
   if (win != NULL)
     {
-      PRINT_DEBUG ("Re-grabbing prefix key\n");
+      PRINT_DEBUG (("Re-grabbing prefix key\n"));
       grab_prefix_key (win->w);
     }
 }
@@ -652,9 +647,7 @@ mapping_notify (XMappingEvent *ev)
   rp_window *cur;
 
   /* Remove the grab on the current prefix key */
-  for (cur = rp_mapped_window_sentinel->next; 
-       cur != rp_mapped_window_sentinel; 
-       cur = cur->next)
+  list_for_each_entry (cur,&rp_mapped_window,node)
     {
       ungrab_prefix_key (cur->w);
     }
@@ -670,9 +663,7 @@ mapping_notify (XMappingEvent *ev)
     }
 
   /* Add the grab on the current prefix key */
-  for (cur = rp_mapped_window_sentinel->next; 
-       cur != rp_mapped_window_sentinel; 
-       cur = cur->next)
+  list_for_each_entry (cur, &rp_mapped_window,node)
     {
       grab_prefix_key (cur->w);
     }
@@ -685,62 +676,62 @@ delegate_event (XEvent *ev)
   switch (ev->type)
     {
     case ConfigureRequest:
-      PRINT_DEBUG ("--- Handling ConfigureRequest ---\n");
+      PRINT_DEBUG (("--- Handling ConfigureRequest ---\n"));
       configure_request (&ev->xconfigurerequest);
       break;
 
     case CreateNotify:
-      PRINT_DEBUG ("--- Handling CreateNotify ---\n");
+      PRINT_DEBUG (("--- Handling CreateNotify ---\n"));
       new_window (&ev->xcreatewindow);
       break;
 
     case DestroyNotify:
-      PRINT_DEBUG ("--- Handling DestroyNotify ---\n");
+      PRINT_DEBUG (("--- Handling DestroyNotify ---\n"));
       destroy_window (&ev->xdestroywindow);
       break;
 
     case ClientMessage:
-      PRINT_DEBUG ("--- Handling ClientMessage ---\n");
+      PRINT_DEBUG (("--- Handling ClientMessage ---\n"));
       client_msg (&ev->xclient);
       break;
 
     case ColormapNotify:
-      PRINT_DEBUG ("--- Handling ColormapNotify ---\n");
+      PRINT_DEBUG (("--- Handling ColormapNotify ---\n"));
       colormap_notify (ev);
       break;
 
     case PropertyNotify:
-      PRINT_DEBUG ("--- Handling PropertyNotify ---\n");
+      PRINT_DEBUG (("--- Handling PropertyNotify ---\n"));
       property_notify (ev);
       break;
 
     case MapRequest:
-      PRINT_DEBUG ("--- Handling MapRequest ---\n");
+      PRINT_DEBUG (("--- Handling MapRequest ---\n"));
       map_request (ev);
       break;
 
     case KeyPress:
-      PRINT_DEBUG ("--- Handling KeyPress ---\n");
+      PRINT_DEBUG (("--- Handling KeyPress ---\n"));
       key_press (ev);
       break;
 
     case UnmapNotify:
-      PRINT_DEBUG ("--- Handling UnmapNotify ---\n");
+      PRINT_DEBUG (("--- Handling UnmapNotify ---\n"));
       unmap_notify (ev);
       break;
 
     case FocusOut:
-      PRINT_DEBUG ("--- Handling FocusOut ---\n");
+      PRINT_DEBUG (("--- Handling FocusOut ---\n"));
       focus_change (&ev->xfocus);
       break;
 
     case FocusIn:
-      PRINT_DEBUG ("--- Handling FocusIn ---\n");
+      PRINT_DEBUG (("--- Handling FocusIn ---\n"));
       focus_change (&ev->xfocus);
       break;
 
     case MappingNotify:
-      PRINT_DEBUG ("--- Handling MappingNotify ---\n");
+      PRINT_DEBUG (("--- Handling MappingNotify ---\n"));
       mapping_notify( &ev->xmapping );
       break;
 
@@ -759,7 +750,7 @@ delegate_event (XEvent *ev)
       break;
 
     default:
-      PRINT_DEBUG ("--- Unknown event %d ---\n",- ev->type);
+      PRINT_DEBUG (("--- Unknown event %d ---\n",- ev->type));
     }
 }
 
@@ -771,7 +762,7 @@ handle_signals ()
     {
       int i;
 
-      PRINT_DEBUG ("Alarm recieved.\n");
+      PRINT_DEBUG (("Alarm recieved.\n"));
 
       /* Only hide the bar if it times out. */
       if (defaults.bar_timeout > 0)
@@ -782,6 +773,29 @@ handle_signals ()
       alarm_signalled = 0;
     }
 
+  if (chld_signalled > 0)
+    {
+      rp_child_info *cur;
+      struct list_head *iter, *tmp;
+
+      /* Report and remove terminated processes. */
+      list_for_each_safe_entry (cur, iter, tmp, &rp_children, node)
+	{
+	  if (cur->terminated)
+	    {
+	      /* Report any child that didn't return 0. */
+	      if (cur->status != 0)
+		marked_message_printf (0,0, " /bin/sh -c \"%s\" finished (%d) ", 
+				       cur->cmd, cur->status);
+	      free (cur->cmd);
+	      free (cur);
+	      list_del  (&cur->node);
+	    }
+	}
+
+      chld_signalled = 0;
+    }
+
   if (hup_signalled > 0)
     {
       clean_up (); 
@@ -790,7 +804,7 @@ handle_signals ()
 
   if (kill_signalled > 0)
     {
-      PRINT_DEBUG ("Exiting\n");
+      PRINT_DEBUG (("Exiting\n"));
       clean_up ();
       exit (EXIT_SUCCESS);
     }
