@@ -32,12 +32,73 @@
 
 #include "ratpoison.h"
 
-const char *unmanaged_window_list[] = {
-#ifdef UNMANAGED_WINDOW_LIST
-  UNMANAGED_WINDOW_LIST,
-#endif
-  NULL
-};
+static char **unmanaged_window_list = NULL;
+static int num_unmanaged_windows = 0;
+
+void
+clear_unmanaged_list ()
+{
+  if (unmanaged_window_list) 
+    {
+      int i;
+
+      for (i = 0; i < num_unmanaged_windows; i++)
+	free(unmanaged_window_list[i]);
+
+      free(unmanaged_window_list);
+
+      unmanaged_window_list = NULL;
+    }
+  num_unmanaged_windows = 0;
+}
+
+char *
+list_unmanaged_windows ()
+{
+  char *tmp = NULL;
+  if (unmanaged_window_list) 
+    {
+      char *tpos;
+      int len = 0;
+      int i;
+
+      for (i = 0; i < num_unmanaged_windows; i++)
+	len += (strlen(unmanaged_window_list[i]) + 1);
+
+      tmp = xmalloc(len + 1);
+      tpos = tmp;
+
+      for (i = 0; i < num_unmanaged_windows; i++) 
+	{
+	  sprintf(tpos, "%s\n", unmanaged_window_list[i]);
+	  tpos += strlen(unmanaged_window_list[i])+1;
+	}
+      tpos--;
+      *tpos = '\0';
+    }
+  return tmp;
+}
+
+void
+add_unmanaged_window (char *name)
+{
+  char **tmp;
+
+  if (!name) return;
+
+  tmp = xmalloc((num_unmanaged_windows + 1) * sizeof(char *));
+
+  if (unmanaged_window_list) 
+    {
+      memcpy(tmp, unmanaged_window_list, num_unmanaged_windows * sizeof(char *));
+      free(unmanaged_window_list);
+    }
+
+  tmp[num_unmanaged_windows] = xstrdup(name);
+  num_unmanaged_windows++;
+
+  unmanaged_window_list = tmp;
+}
 
 extern Atom wm_state;
 
@@ -61,10 +122,19 @@ ungrab_prefix_key (Window w)
 #endif
 }
 
-screen_info*
+rp_screen*
 current_screen ()
 {
-  return &screens[rp_current_screen];
+  int i;
+
+  for (i=0; i<num_screens; i++)
+    {
+      if (screens[i].screen_num == rp_current_screen)
+	return &screens[i];
+    }
+  
+  /* This should never happen. */
+  return &screens[0];
 }
 
 void
@@ -77,13 +147,13 @@ update_normal_hints (rp_window *win)
   /* Print debugging output for window hints. */
 #ifdef DEBUG
   if (win->hints->flags & PMinSize)
-    PRINT_DEBUG ("minx: %d miny: %d\n", win->hints->min_width, win->hints->min_height);
+    PRINT_DEBUG (("minx: %d miny: %d\n", win->hints->min_width, win->hints->min_height));
 
   if (win->hints->flags & PMaxSize)
-    PRINT_DEBUG ("maxx: %d maxy: %d\n", win->hints->max_width, win->hints->max_height);
+    PRINT_DEBUG (("maxx: %d maxy: %d\n", win->hints->max_width, win->hints->max_height));
 
   if (win->hints->flags & PResizeInc)
-    PRINT_DEBUG ("incx: %d incy: %d\n", win->hints->width_inc, win->hints->height_inc);
+    PRINT_DEBUG (("incx: %d incy: %d\n", win->hints->width_inc, win->hints->height_inc));
 
 #endif
 }
@@ -92,37 +162,42 @@ update_normal_hints (rp_window *win)
 static char *
 get_wmname (Window w)
 {
-  char *name;
-  XTextProperty text;
-  char **name_list;
-  int list_len;
+  Atom actual_type;
+  int actual_format;
+  int status;
+  unsigned long n;
+  unsigned long bytes_after;
+  unsigned char *name = NULL;
+  char *ret;
 
-  if (!XGetWMName (dpy, w, &text))
+  status = XGetWindowProperty (dpy, w, wm_name, 0L, 100L, False, 
+			       XA_STRING, &actual_type, &actual_format, 
+			       &n, &bytes_after, &name);
+
+  PRINT_DEBUG (("XGetWindowProperty: %d %ld %d %ld %ld '%s'\n", status, actual_type, 
+		actual_format, n, bytes_after, name));
+
+  if (status != Success || name == NULL)
     {
-      PRINT_DEBUG ("I can't get the WMName.\n");
+      PRINT_DEBUG (("I can't get the WMName.\n"));
       return NULL;
     }
 
-  if (!XTextPropertyToStringList (&text, &name_list, &list_len))
+  if (n == 0)
     {
-      PRINT_DEBUG ("Error retrieving TextList.\n");
+      PRINT_DEBUG (("I can't get the WMName.\n"));
+      XFree (name);
       return NULL;
     }
 
-  if (list_len > 0)
-    {
-      name = xmalloc (strlen (name_list[0]) + 1);
-      strcpy (name, name_list[0]);
-    }
-  else
-    {
-      name = NULL;
-    }
+  PRINT_DEBUG (("WM_NAME: '%s'\n", name));
 
-  /* Its our responsibility to free this. */ 
-  XFreeStringList (name_list);
+  /* duplicate the string into our own buffer, and free the one given
+     to us by X. */
+  ret = xstrdup (name);
+  XFree (name);
 
-  return name;
+  return ret;
 }
 
 static XClassHint *
@@ -134,7 +209,7 @@ get_class_hints (Window w)
 
   if (class == NULL)
     {
-      PRINT_ERROR ("Not enough memory for WM_CLASS structure.\n");
+      PRINT_ERROR (("Not enough memory for WM_CLASS structure.\n"));
       exit (EXIT_FAILURE);
     }
 
@@ -199,9 +274,6 @@ update_window_name (rp_window *win)
 {
   char *newstr;
 
-  /* Don't overwrite the window name if the user specified one. */
-/*   if (win->named) return 0; */
-
   newstr = get_wmname (win->w);
   if (newstr != NULL)
     {
@@ -274,14 +346,13 @@ update_window_information (rp_window *win)
 void
 unmanage (rp_window *w)
 {
-  return_window_number (w->number);
-  remove_from_list (w);
-
-  free_window (w);
+  numset_release (rp_window_numset, w->number);
+  list_del (&w->node);
+  free_window (w);  
 
 #ifdef AUTO_CLOSE
-  if (rp_mapped_window_sentinel->next == rp_mapped_window_sentinel
-      && rp_mapped_window_sentinel->prev == rp_mapped_window_sentinel)
+  if (rp_mapped_window->next == &rp_mapped_window
+      && rp_mapped_window->prev == &rp_mapped_window)
     {
       /* If the mapped window list is empty then we have run out of
  	 managed windows, so kill ratpoison. */
@@ -298,7 +369,7 @@ unmanage (rp_window *w)
 
 /* When starting up scan existing windows and start managing them. */
 void
-scanwins(screen_info *s)
+scanwins(rp_screen *s)
 {
   rp_window *win;
   XWindowAttributes attr;
@@ -306,7 +377,7 @@ scanwins(screen_info *s)
   Window dw1, dw2, *wins;
 
   XQueryTree(dpy, s->root, &dw1, &dw2, &wins, &nwins);
-  PRINT_DEBUG ("windows: %d\n", nwins);
+  PRINT_DEBUG (("windows: %d\n", nwins));
 
   for (i = 0; i < nwins; i++) 
     {
@@ -321,11 +392,16 @@ scanwins(screen_info *s)
 
       win = add_to_window_list (s, wins[i]);
 
-      PRINT_DEBUG ("map_state: %d\n", attr.map_state);
+      PRINT_DEBUG (("map_state: %s\n", 
+		    attr.map_state == IsViewable ? "IsViewable":
+		    attr.map_state == IsUnviewable ? "IsUnviewable" : "IsUnmapped"));
+      PRINT_DEBUG (("state: %s\n", 
+		    get_state(win) == IconicState ? "Iconic":
+		    get_state(win) == NormalState ? "Normal" : "Other"));
       
       /* Collect mapped and iconized windows. */
       if (attr.map_state == IsViewable
-	  || (attr.map_state == IsUnmapped 
+	  || (attr.map_state == IsUnmapped
 	      && get_state (win) == IconicState))
 	map_window (win);
     }
@@ -339,18 +415,19 @@ unmanaged_window (Window w)
   char *wname;
   int i;
 
-  for (i = 0; unmanaged_window_list[i]; i++) 
+  if (!unmanaged_window_list) return 0;
+
+  for (i = 0; i < num_unmanaged_windows; i++) 
     {
-      wname = get_wmname (w);
-      if (!wname)
-	return 0;
-      if (!strcmp (unmanaged_window_list[i], wname))
+      wname = get_wmname(w);
+      if (!wname) return 0;
+      if (!strcmp(unmanaged_window_list[i], wname)) 
 	{
-	  free (wname);
+	  free(wname);
 	  return 1;
 	}
+      free(wname);
     }
-
   return 0;
 }
 
@@ -398,26 +475,30 @@ get_state (rp_window *win)
 static void
 move_window (rp_window *win)
 {
-  if (win->frame == NULL) 
+  rp_frame *frame;
+
+  if (win->frame_number == EMPTY) 
     return;
 
+  frame = win_get_frame (win);
+
   /* X coord. */
   switch (win->gravity)
     {
     case NorthWestGravity:
     case WestGravity:
     case SouthWestGravity:
-      win->x = win->frame->x;
+      win->x = frame->x;
       break;
     case NorthGravity:
     case CenterGravity:
     case SouthGravity:
-      win->x = win->frame->x + (win->frame->width - win->border * 2) / 2 - win->width / 2;
+      win->x = frame->x + (frame->width - win->border * 2) / 2 - win->width / 2;
       break;
     case NorthEastGravity:
     case EastGravity:
     case SouthEastGravity:
-      win->x = win->frame->x + win->frame->width - win->width - win->border;
+      win->x = frame->x + frame->width - win->width - win->border;
       break;
     }
 
@@ -427,17 +508,17 @@ move_window (rp_window *win)
     case NorthEastGravity:
     case NorthGravity:
     case NorthWestGravity:
-      win->y = win->frame->y;
+      win->y = frame->y;
       break;
     case EastGravity:
     case CenterGravity:
     case WestGravity:
-      win->y = win->frame->y + (win->frame->height - win->border * 2) / 2 - win->height / 2;
+      win->y = frame->y + (frame->height - win->border * 2) / 2 - win->height / 2;
       break;
     case SouthEastGravity:
     case SouthGravity:
     case SouthWestGravity:
-      win->y = win->frame->y + win->frame->height - win->height - win->border;
+      win->y = frame->y + frame->height - win->height - win->border;
       break;
     }
 }
@@ -447,18 +528,18 @@ move_window (rp_window *win)
 static void
 maximize_transient (rp_window *win)
 {
-  rp_window_frame *frame;
+  rp_frame *frame;
   int maxx, maxy;
 
+  frame = win_get_frame (win);
+
   /* We can't maximize a window if it has no frame. */
-  if (win->frame == NULL)
+  if (frame == NULL)
     return;
 
   /* Set the window's border */
   win->border = defaults.window_border_width;
 
-  frame = win->frame;
-
   /* Always use the window's current width and height for
      transients. */
   maxx = win->width;
@@ -467,8 +548,8 @@ maximize_transient (rp_window *win)
   /* Fit the window inside its frame (if it has one) */
   if (frame)
     {
-      PRINT_DEBUG ("frame width=%d height=%d\n", 
-		   frame->width, frame->height);
+      PRINT_DEBUG (("frame width=%d height=%d\n", 
+		   frame->width, frame->height));
 
       if (maxx + win->border * 2 > frame->width) maxx = frame->width - win->border * 2;
       if (maxy + win->border * 2 > frame->height) maxy = frame->height - win->border * 2;
@@ -494,7 +575,7 @@ maximize_transient (rp_window *win)
       maxy = amount + win->height;
     }
 
-  PRINT_DEBUG ("maxsize: %d %d\n", maxx, maxy);
+  PRINT_DEBUG (("maxsize: %d %d\n", maxx, maxy));
 
   win->width = maxx;
   win->height = maxy;
@@ -505,18 +586,18 @@ maximize_transient (rp_window *win)
 static void
 maximize_normal (rp_window *win)
 {
-  rp_window_frame *frame;
+  rp_frame *frame;
   int maxx, maxy;
 
+  frame = win_get_frame (win);
+
   /* We can't maximize a window if it has no frame. */
-  if (win->frame == NULL)
+  if (frame == NULL)
     return;
 
   /* Set the window's border */
   win->border = defaults.window_border_width;
 
-  frame = win->frame;
-
   /* Honour the window's maximum size */
   if (win->hints->flags & PMaxSize)
     {
@@ -532,8 +613,8 @@ maximize_normal (rp_window *win)
   /* Fit the window inside its frame (if it has one) */
   if (frame)
     {
-      PRINT_DEBUG ("frame width=%d height=%d\n", 
-		   frame->width, frame->height);
+      PRINT_DEBUG (("frame width=%d height=%d\n", 
+		   frame->width, frame->height));
 
       if (maxx > frame->width) maxx = frame->width - win->border * 2;
       if (maxy > frame->height) maxy = frame->height - win->border * 2;
@@ -559,7 +640,7 @@ maximize_normal (rp_window *win)
       maxy = amount + win->height;
     }
 
-  PRINT_DEBUG ("maxsize: %d %d\n", maxx, maxy);
+  PRINT_DEBUG (("maxsize: %d %d\n", maxx, maxy));
 
   win->width = maxx;
   win->height = maxy;
@@ -582,8 +663,8 @@ maximize (rp_window *win)
   /* Reposition the window. */
   move_window (win);
 
-  PRINT_DEBUG ("Resizing window '%s' to x:%d y:%d w:%d h:%d\n", window_name (win), 
-	       win->x, win->y, win->width, win->height);
+  PRINT_DEBUG (("Resizing window '%s' to x:%d y:%d w:%d h:%d\n", window_name (win), 
+	       win->x, win->y, win->width, win->height));
 
 
   /* Actually do the maximizing. */
@@ -632,16 +713,16 @@ force_maximize (rp_window *win)
 void
 map_window (rp_window *win)
 {
-  PRINT_DEBUG ("Mapping the unmapped window %s\n", window_name (win));
+  PRINT_DEBUG (("Mapping the unmapped window %s\n", window_name (win)));
 
   /* Fill in the necessary data about the window */
   update_window_information (win);
-  win->number = get_unique_window_number ();
+  win->number = numset_request (rp_window_numset);
   grab_prefix_key (win->w);
 
   /* Put win in the mapped window list */
-  remove_from_list (win);
-  insert_into_list (win, rp_mapped_window_sentinel);
+  list_del (&win->node);
+  insert_into_list (win, &rp_mapped_window); 
 
   /* The window has never been accessed since it was brought back from
      the Withdrawn state. */
@@ -671,7 +752,7 @@ hide_window (rp_window *win)
   if (win == NULL) return;
 
   /* An unmapped window is not inside a frame. */
-  win->frame = NULL;
+  win->frame_number = EMPTY;
 
   /* Ignore the unmap_notify event. */
   XSelectInput(dpy, win->w, WIN_EVENTS&~(StructureNotifyMask));
@@ -716,15 +797,14 @@ withdraw_window (rp_window *win)
 {
   if (win == NULL) return;
 
-  PRINT_DEBUG ("withdraw_window on '%s'\n", window_name (win));
+  PRINT_DEBUG (("withdraw_window on '%s'\n", window_name (win)));
 
   /* Give back the window number. the window will get another one,
      if it is remapped. */
-  return_window_number (win->number);
+  numset_release (rp_window_numset, win->number);
   win->number = -1;
 
-  remove_from_list (win);
-  append_to_list (win, rp_unmapped_window_sentinel);
+  list_move_tail(&win->node, &rp_unmapped_window);
 
   ignore_badwindow++;
 
@@ -739,20 +819,18 @@ withdraw_window (rp_window *win)
 void
 hide_others (rp_window *win)
 {
-  rp_window_frame *frame;
+  rp_frame *frame;
   rp_window *cur;
 
   if (win == NULL) return;
   frame = find_windows_frame (win);
   if (frame == NULL) return;
 
-  for (cur = rp_mapped_window_sentinel->next; 
-       cur != rp_mapped_window_sentinel; 
-       cur = cur->next)
+  list_for_each_entry (cur, &rp_mapped_window, node)
     {
       if (find_windows_frame (cur) 
 	  || cur->state != NormalState 
-	  || cur->frame != frame)
+	  || cur->frame_number != frame->number)
 	continue;
 
       hide_window (cur);
